from argparse import ArgumentParser
from cast_arg.convert import GeneratePPT
from cast_arg.config import Config
# from convert import GeneratePPT
# from config import Config

__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2023, CAST Software"

if __name__ == '__main__':
    print('\nCAST Assessment Deck Generation Tool (ARG)')
    print('Copyright (c) 2023 CAST Software Inc.\n')
    print('If you need assistance, please contact oneclick@castsoftware.com\n')

    parser = ArgumentParser(description='Assessment Report Generation Tool')
    parser.add_argument('-c','--config', required=True, help='Configuration properties file')
    args = parser.parse_args()
    ppt = GeneratePPT(Config(args.config))
    ppt.save_ppt()




"""
    Read and validate configuration file
"""
from cast_common.logger import Logger,DEBUG, INFO, WARN, ERROR
from json import load
from argparse import ArgumentParser
from json import JSONDecodeError

__author__ = "Nevin Kaplan"
__copyright__ = "Copyright 2022, CAST Software"
__email__ = "n.kaplan@castsoftware.com"

class Config():
    log = None
    log_translate = {} 
    def __init__(self, config:str):
        #super().__init__("Config")

        self.log_translate['info']=INFO
        self.log_translate['warn']=WARN
        self.log_translate['error']=ERROR
        self.log_translate['debug']=DEBUG

        #do all required fields contain data
        try:
            with open(config, 'rb') as config_file:
                self._config = load(config_file)

            #get logging configuration
            # is there a logging, if not add it now?
            if 'logging' not in self._config:
                self._config['logging']={}

            # if any logging entries are missing, add them now   
            log_config = self._config['logging']
            for v in ['config','aip','highlight','generate']:
                if v not in log_config:
                   log_config[v]='info'

            # convert the entries into something the logger can understand
            for idx,value in enumerate(log_config):
                log_config[value]=self.log_translate[log_config[value]]
            self.log=Logger('config',log_config['config'])

            for v in ['project','application','company','template']:
                self.log.debug(f'Validating {v}')
                if v not in self._config or len(self._config[v]) == 0:
                    raise ValueError(f"Required field '{v}' is missing from config.json")

            apps = self._config['application']
            for idx,value in enumerate(apps):
                if 'aip' in value: 
                    if len(value['aip']) == 0:
                        raise ValueError(f'No AIP id for application {idx+1}')
                    aip=value['aip']
                
                #title and highlight values are both optional
                #if they are not filled in then populate with the aip value
                if 'highlight' not in value or len(value['highlight'])==0:
                    value['highlight']=aip
                if 'title' not in value or len(value['title'])==0:
                    value['title']=aip

            if 'rest' not in self._config:
                raise ValueError(f"Required field 'rest' is missing from config.json")

            for v in ['AIP','Highlight']:
                if v not in self._config['rest']:
                    raise ValueError(f"Required field '{v}' is missing from config.json")

            self.__rest_settings(self._config['rest']['AIP'])
            self.__rest_settings(self._config['rest']['Highlight'])
            if 'instance' not in self._config['rest']['Highlight']:
                raise ValueError(f"Required field 'instance' is missing from config.json")

        except JSONDecodeError as e:
            msg = str(e)
            self.error('Configuration file must be in a JSON format')
            exit()

        except ValueError as e:
            msg = str(e)
            self.log.error(msg)
            exit()

    def __rest_settings(self,dict):
        for v in ["Active","URL","user","password"]:
            if v not in dict:
                raise ValueError(f"Required field '{v}' is missing from config.json")

    @property
    def project(self):
        return self._config['project']

    @property
    def company(self):
        return self._config['company']

    @property
    def template(self):
        return self._config['template']

    @property
    def output(self):
        return self._config['output']

    @property
    def cause(self):
        return self._config['cause']

    @property
    def application(self):
        return self._config['application']

    @property
    def aip_list(self):
        rslt = []
        for app in self.application:
            rslt.append(app['aip'])
        return rslt

    @property
    def hl_list(self):
        rslt = []
        for app in self.application:
            rslt.append(app['highlight'])
        return rslt

    @property
    def title_list(self):
        rslt = []
        for app in self.application:
            rslt.append(app['title'])
        return rslt

    @property
    def logging_aip(self):
        return self._config['logging']['aip']
    @property
    def logging_highlight(self):
        return self._config['logging']['highlight']
    @property
    def logging_generate(self):
        return self._config['logging']['generate']

    def aip_name(self,idx):
        return self.application[idx]['aip']

    def hl_name(self,idx):
        return self.application[idx]['highlight']

    def title(self,idx):
        return self.application[idx]['title']

    @property
    def rest(self):
        return self._config['rest']

    @property
    def aip_active(self):
        return self.rest['AIP']['Active']

    @property
    def aip_url(self):
        return self.rest['AIP']['URL']

    @property
    def aip_user(self):
        return self.rest['AIP']['user']

    @property
    def aip_password(self):
        return self.rest['AIP']['password']

    @property
    def hl_active(self):
        return self.rest['Highlight']['Active']

    @property
    def hl_url(self):
        return self.rest['Highlight']['URL']

    @property
    def hl_user(self):
        return self.rest['Highlight']['user']

    @property
    def hl_password(self):
        return self.rest['Highlight']['password']

    @property
    def hl_instance(self):
        return self.rest['Highlight']['instance']




"""
parser = ArgumentParser(description='CAST Assessment Report Generation (ARG)')
parser.add_argument('-c','--config', required=True, help='Configuration properties file')
args = parser.parse_args()

c = ARGConfig(args.config)
print (f'app 1: {c.aip_app(0)}')
"""



from cast_common.powerpoint import PowerPoint as common_ppt
from cast_common.logger import INFO
from pptx.chart.data import CategoryChartData
from pandas import Series

__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2022, CAST Software"


class PowerPoint(common_ppt):
    """
    This module holds techniques explicitly designed for 
    constructing PowerPoint slides. The majority of its capabilities 
    come from the PowerPoint module in the com.castsoftware.uc.python.common library. 

    Args:
        common_ppt: PowerPoint class found in com.castsoftware.uc.python.common
    """

    def __init__(self,template:str=None,output:str=None,log_level=INFO) -> None:
        super().__init__(template,output,log_level)

    def replace_risk_factor(self, grades:Series, app_no:int=0, search_str:str=None):
        if search_str == None:
            search_str=f'{{app{app_no}_risk_'
        
        for slide in self._prs.slides:
            for shape in slide.shapes:
                if shape.has_text_frame:
                    for paragraph in shape.text_frame.paragraphs:
                        if paragraph.text.find(search_str)!=-1:
                            run = self._merge_runs(paragraph)
                            # cur_text=''
                            # first=True
                            # for run in paragraph.runs:
                            #     cur_text = cur_text + run.text
                            #     if first != True:
                            #         self.delete_run(run)
                            #     first=False
                            # run = paragraph.runs[0]
                            # run.text = cur_text

                            while True:
                                t = cur_text.find(search_str)
                                if t == -1: 
                                    break
                                g = cur_text[t+len(search_str):]
                                g = g[:g.find("}")]
                                if g:
                                    try:
                                        grade=grades[g]
                                        risk = ''
                                        if grade < 2:
                                            risk = 'high'
                                        elif grade < 3:
                                            risk = 'medium'
                                        else:
                                            risk = 'low'

                                        cur_text = cur_text.replace(f'{search_str}{g}}}',risk)
                                        run.text = cur_text
                                    except KeyError:
                                        self.debug(f'invalid key: {g}')
                                        break;

    def update_grade_slider(self, shape,data):
        #shape = self.get_shape_by_name(name)
        try:
            if shape.has_chart:
                chart_data = CategoryChartData()
                # chart_data.categories = titles
                
                chart_data.categories = ['grade']
                chart_data.add_series('Series 1', data)
                shape.chart.replace_data(chart_data)
        except AttributeError as ae:
            self.warning(f'Attribute Error, invalid template configuration: {shape.name} ({ae})')
        except KeyError as ke:
            self.warning(f'Key Error, invalid template configuration: {shape.name} ({ke})')
        except TypeError as t:
            self.warning(f'Type Error, invalid template configuration: {shape.name} ({t})')

    def remove_empty_placeholders(self):
        for slide in self._prs.slides:
            for placeholder in slide.shapes.placeholders:
                if placeholder.has_text_frame and placeholder.text_frame.text == "":
                    sp = placeholder._sp
                    sp.getparent().remove(sp)

    def replace_loc(self, loc, app_no):
        loc_short = "{0:,.0f} LoC".format(loc) 
        if loc > 1000000:
            loc_short = "~{0:,.2f} MLoC".format(loc/1000000) 
        elif loc < 1000000 and loc > 1000:
            loc_short = "~{0:,.0f} KLoC".format(loc/1000) 
        self.replace_text(f'{{app{app_no}_loc}}',f'{loc:,.0f}')
        self.replace_text(f'{{app{app_no}_loc_short}}',loc_short)

        size_catagory = 'small'
        if loc > 1000000:
            size_catagory = 'very large'
        elif loc > 500000:
            size_catagory = 'large'
        self.replace_text(f'{{app{app_no}_loc_category}}',size_catagory)




from IPython.display import display
from cast_common.highlight import Highlight
from cast_common.logger import Logger, INFO,DEBUG
from cast_common.powerpoint import PowerPoint
from cast_common.util import format_table
from pandas import DataFrame,Series,json_normalize,ExcelWriter
from os.path import abspath
from sys import exc_info

class GreenIt(Highlight):


    def __init__(self,output:str,  
                 hl_user:str=None, hl_pswd:str=None,hl_basic_auth=None, hl_instance:int=0,
                 hl_apps:str=[],hl_tags:str=[], 
                 hl_base_url:str=None, 
                 log_level=INFO, timer_on=False):
        super().__init__(hl_user, hl_pswd,hl_basic_auth, hl_instance,hl_apps,hl_tags, hl_base_url, log_level, timer_on)
        self._output = output

    def report(self,app:str,app_no:int) -> bool:
        status = True
        try:
            index = self.get_green_indexes(app)
            for idx,val in index.items():
                if idx == 'greenOccurrences':
                    val = int(val)
                tag = f'{{app{app_no}_{idx}}}'
                PowerPoint.ppt.replace_text(tag,val)

            detail = self.get_green_detail(app)
            detail = detail[detail['Occurrences'] != 0]
            detail=detail.drop(columns=['Contribution'])

            agr = detail[['Technology','Occurrences']].groupby('Technology').aggregate('sum').reindex()
            PowerPoint.ppt.update_chart(f'app{app_no}_GreenTechPieChart',agr)
            agr.sort_values('Occurrences',ascending=False,inplace=True)
            PowerPoint.ppt.replace_text(f'{{app{app_no}_green_top_lang}}',agr.index[0])
            PowerPoint.ppt.replace_text(f'{{app{app_no}_green_top_lang_count}}',int(agr.iloc[0,0]))

            agr = detail[['Name','Occurrences']].groupby('Name').aggregate('sum').reindex()
            agr.sort_values('Occurrences',ascending=False,inplace=True)
            PowerPoint.ppt.replace_text(f'{{app{app_no}_green_top_rule1}}',agr.index[0])
            PowerPoint.ppt.replace_text(f'{{app{app_no}_green_top_rule2}}',agr.index[1])

            detail.sort_values(by=['Occurrences'],ascending=False,inplace=True)
            #detail['Contribution'] = detail['Contribution'].apply(lambda x: '{0:.2f}%'.format(x).rjust(10))
            detail['Occurrences'] = detail['Occurrences'].apply(lambda x: '{0:,.0f}'.format(x).rjust(10))
            self.create_excel(app,detail,self._output)

            #detail = detail.astype({'Contribution':'string'})
            detail['Technology'] = detail['Technology'].str.ljust(20)
            PowerPoint.ppt.update_table(f'app{app_no}_GreenDetailTable',detail,include_index=False,max_rows=8)


            pass      
        except Exception as ex:
            ex_type, ex_value, ex_traceback = exc_info()
            self.log.error(f'{ex_type.__name__}: {ex_value}')
            status = False

        return status

    def get_green_indexes(self,app_name:str) -> Series:
        self.log.info(f'Retrieving green it index data for: {app_name}')
        try:
            df = json_normalize(self._get_metrics(app_name)['greenDetail'],['greenIndexDetails'],meta=['technology','greenIndexScan']).dropna(axis='columns')
            df = df[['greenIndexScan','greenOccurrences','greenEffort']]
            df = df.aggregate(['sum','average'])

            if self.log.is_debug:
                self.log.debug('aggragation')
                display(df)

            rslt = Series()
            rslt.loc['greenEffort']=round(df.loc['sum','greenEffort']/60/8,1)
            rslt.loc['greenOccurrences']=round(df.loc['sum','greenOccurrences'],0)
            rslt.loc['greenIndexScan']=round(df.loc['average','greenIndexScan']*100,1)

            if self.log.is_debug:
                self.log.debug('final')
                display(rslt)

            return rslt
        except KeyError as ke:
            self.warning(f'{app_name} has no Green IT Data')
            return None

    def get_green_detail(self,app_name:str)->DataFrame:
        """Highlight green it data

        Args:
            app_name (str): name of the application

        Returns:
            DataFrame: flattened version of the Highlight green it data
        """
        self.log.info(f'Retrieving green it detail data for: {app_name}')
        try:
            df = json_normalize(self._get_metrics(app_name)['greenDetail'],['greenIndexDetails'],meta=['technology','greenIndexScan']).dropna(axis='columns')
            df.drop(columns=['greenRequirement.id','greenRequirement.hrefDoc','triggered','greenRequirement.ruleType','greenIndexScan'],inplace=True)
            df.rename(columns={'contributionScore':'Contribution',
                               'greenOccurrences':'Occurrences',
                               'greenRequirement.display':'Name',
                               'greenEffort':'Effort',
                               'technology':'Technology'},
                               inplace=True)
            df['Effort']=df['Effort'].div(60).div(8).round(2)
            df['Contribution']=df['Contribution'].round(2)
            return df[['Name','Technology','Contribution','Occurrences','Effort']]
        except KeyError as ke:
            self.warning(f'{app_name} has no Green IT Data')
            return None

    def create_excel(self,app_name:str,data:DataFrame,output:str):
        file_name = abspath(f'{output}/greenIt-Reporting-{app_name}.xlsx')
        writer = ExcelWriter(file_name, engine='xlsxwriter')
        format_table(writer,data,'Detail',width=[75,25,15,15,15],total_line=True)
        writer.close()








from cast_common.logger import Logger, INFO,DEBUG
from cast_common.aipRestCall import AipRestCall
from cast_common.mri import MRI
from cast_arg.powerpoint import PowerPoint

from pandas import json_normalize,DataFrame
from tqdm import tqdm



class TechDetailTable(AipRestCall):

    def report(self,app_name:str,app_no:int) -> bool:
        # app_tag = f'app{app_no}'

        sizing = {
        '10151':'Number of Code Lines' #, 
        #'67011':'Critical Violations'
        }

        domain_id = self.get_domain(app_name)
        snapshot = self.get_latest_snapshot(domain_id)
        sizing_df = self.get_sizing_by_technology(domain_id,snapshot,sizing)
        (ap_df,ap_summary_df)=self.get_action_plan(domain_id,snapshot['id'])        

        sizing_df['Fix Now']=0
        sizing_df['Near Term']=0
        sizing_df['Mid Term']=0
        # sizing_df['Long Term']=0
        for key, value in sizing_df.iterrows():
            if key=='All':
                sizing_df.at[key,'Fix Now'] = len(ap_df[ap_df['Action Plan Priority']=='Fix Now'])
                sizing_df.at[key,'Near Term'] = len(ap_df[ap_df['Action Plan Priority']=='Near Term'])
                sizing_df.at[key,'Mid Term'] = len(ap_df[ap_df['Action Plan Priority']=='Mid Term'])
                # sizing_df.at[key,'Long Term'] = len(ap_df[ap_df['Action Plan Priority']=='Long Term'])
            else:
                sizing_df.at[key,'Fix Now'] = self._get_counts(ap_df,'Fix Now',key)
                sizing_df.at[key,'Near Term'] = self._get_counts(ap_df,'Near Term',key)
                sizing_df.at[key,'Mid Term'] = self._get_counts(ap_df,'Mid Term',key)
                # sizing_df.at[key,'Long Term'] = self._get_counts(ap_df,'Long Term',key)
            pass

        sizing_df = sizing_df.astype('int')

        # sizing_df=sizing_df[['Number of Code Lines','Fix Now','Near Term','Mid Term','Long Term','Critical Violations']]
        sizing_df['Number of Code Lines'] = sizing_df['Number of Code Lines'].map('{:,.0f}'.format)
        PowerPoint.ppt.update_table(f'app{app_no}_technical_details_table',sizing_df,header_rows=2)
        pass       

    def _get_counts(self,data:DataFrame,priority:str,tech:str) -> int:
        fn = data[data['Action Plan Priority']==priority]
        fn = fn[fn['Technology'].str.lower().str.contains(tech.lower())]
        return len(fn)
        


        pass

    # def replace_text(self,prs, prefix, item, data):
    #     tag = f'{{{prefix}_{item}}}'
    #     self.log.debug(f'{tag}: {data}')
    #     prs.replace_text(tag,data)




from io import RawIOBase

from pandas import DataFrame,json_normalize,concat,notnull
from pptx.chart.data import Category
from IPython.display import display

from time import perf_counter, ctime
from copy import copy
from cast_common.logger import Logger,DEBUG, INFO, ERROR
from cast_common.restAPI import RestCall
from cast_common.aipRestCall import AipRestCall
from cast_common.hlRestCall import HLRestCall
from cast_common.powerpoint import PowerPoint

import enum
import urllib.parse
import numpy as np 
import json

__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2022, CAST Software"

"""
    This class is used to retrieve information from the CAST AIP REST API
"""
class AipData(AipRestCall):
    _data={}
    _base=[]

    _sizing = {
       '10151':'Number of Code Lines', 
       '10107':'Number of Comment Lines', 
       '10109':'Number of Commented-out Code Lines',
       '67011':'Critical Violations'
    }

    _tech_sizing = {
        '10151':'LoC',
        '10154':'Files',
        '10155':'Classes',
        '10158':'SQL Artifacts',
        '10163':'Tables'
    }

    _imp_list = ['66070','61001','61013','61004','66009',
                 '61009','61029','61031','61010','61011',
                 '61026','61027','66068','61018','61019',
                 '61014','61015','61020','61003','61024',
                 '66069','66063','66066','66062','66065',
                 '66064']

    _doc_list = ['61008','61007','61017','61017','61006','61028','61022','61023']

    _health_grade_ids = ['Efficiency','Robustness','Security','Changeability','Transferability']

    def __init__(self, config,output:str=None,timer_on=False,log_level=INFO):
        super().__init__(base_url=config.aip_url, user=config.aip_user, pswd=config.aip_password, 
                         track_time=timer_on,log_level=log_level)

        #self._rest=rest
        #self._base=app_list

        for s in config.aip_list:
            self.info(f'*******Collecting AIP data for {s}********')
            self._data[s]={}
            self._data[s]['has data'] = False
#            central_schema = f'{s}_central'.replace('-','_')
            domain_id = self.get_domain(s)
            if domain_id == -1:
                raise SystemExit  #connection failed, exit here
            if domain_id is not None:
                self._data[s]['domain_id']=domain_id
                snapshot = self.get_latest_snapshot(domain_id)
                self._data[s]['snapshot']=snapshot
                if self._data[s]['snapshot']:
                    self.info('AIP is active')
                    self._data[s]['has data'] = True

                    self.info('action plan data')
                    (ap_df,ap_summary_df) = self.get_action_plan(domain_id,self._data[s]['snapshot']['id']) 
                    self._data[s]['action_plan']=ap_df
                    self._data[s]['action_plan_summary']=ap_summary_df

                    self.info('TQI compliance data')
                    self._data[s]['tqi_compliance']=self.aggregate_violation_ratio(domain_id,snapshot['id'],'60017',self._imp_list)

                    self.info('documentation compliance data')
                    self._data[s]['doc_compliance']=self.aggregate_violation_ratio(domain_id,snapshot['id'],'66033',self._doc_list,False)

                    self.info('Grades by technololgy')
                    self._data[s]['grades']=self.get_grades_by_technology(domain_id,self._data[s]['snapshot'])

                    self.info('Sizing by technical')
                    self._data[s]['sizing']=self.get_sizing_by_technology(domain_id,self._data[s]['snapshot'],self._sizing)

                    self.info('LOC data')
                    self._data[s]['loc_sizing']=self.get_sizing(domain_id,self._sizing) 

                    self.info('Technical sizing data')
                    self._data[s]['tech_sizing']=self.get_sizing(domain_id,self._tech_sizing) 

                    self.info('Violation sizing data')
                    self._data[s]['violation_sizing']=self.get_violation_CR(domain_id)

                    self.info('Critical rules')
                    self._data[s]['critical_rules']=self.get_rules(domain_id,self._data[s]['snapshot']['id'],60017,non_critical=False)

                    self.info('ISO rules')
                    self._data[s]['ISO']=self._get_iso_rules(domain_id,snapshot['id'])

                    # self.info('Green IT Index')
                    # self._data[s]['GreenIT']=self._get_green_rules(domain_id,snapshot['id'])



                    self.info('AIP data retrieval complete')
            else:
                self.logger.warn(f'Domain not found for {s}')

    def has_data(self, app):
        return self._data[app]['has data']

    def data(self,app):
        return self._data[app]

    def iso_rules(self, app):
        return self._data[app]['ISO']

    def _get_green_it_rules(self,domain_id,snapshot_id):
        iso={20140522:"Green IT Index"}

        rslt_df = DataFrame()
        rslt_df.style.set_properties(subset=['text'],**{'text-align': 'left'})
        for key, value in iso.items():
            try:
                temp = DataFrame(columns=['category','violation'])
                rp = json_normalize(self.get_rules(domain_id,snapshot_id,key)['rulePattern'])
                temp['violation'] = rp['name']
                temp['category'] = value
                temp = temp.groupby(['category','violation']).size().reset_index(name='count') 

                total = temp.groupby(['category'])['count'].sum().reset_index(name='count') 
                total['violation']=''
                total = total[['category','violation','count']]

                rslt_df = concat([rslt_df,total,temp])
            except KeyError as e:
                self.warning(f'no iso rules for {value} ({e})')

        if 'category' in rslt_df.columns:
            rslt_df['category'] = rslt_df['category'].mask(rslt_df['category'].ne(rslt_df['category'].shift()).cumsum().duplicated(), '')
        else:
            self.warning('No ISO rules found')
            
        return rslt_df

    def _get_iso_rules(self,domain_id,snapshot_id):
        iso={1061004:"Security",
             1061003:"Reliabllity",
             1061002:"Performance-Efficiency",
             1061001:"Maintainability",
        }

        rslt_df = DataFrame()
        rslt_df.style.set_properties(subset=['text'],**{'text-align': 'left'})
        for key, value in iso.items():
            try:
                temp = DataFrame(columns=['category','violation'])
                rp = json_normalize(self.get_rules(domain_id,snapshot_id,key,return_raw=True)['rulePattern'])
                temp['violation'] = rp['name']
                temp['category'] = value
                temp = temp.groupby(['category','violation']).size().reset_index(name='count') 

                total = temp.groupby(['category'])['count'].sum().reset_index(name='count') 
                total['violation']=''
                total = total[['category','violation','count']]

                rslt_df = concat([rslt_df,total,temp])
            except KeyError as e:
                self.warning(f'no iso rules for {value} ({e})')

        if 'category' in rslt_df.columns:
            rslt_df['category'] = rslt_df['category'].mask(rslt_df['category'].ne(rslt_df['category'].shift()).cumsum().duplicated(), '')
        else:
            self.warning('No ISO rules found')
            
        return rslt_df

    def domain(self, app):
        return self.data(app)['domain_id']

    def snapshot(self, app):
        return self.data(app)['snapshot']

    def grades(self, app):
        return self.data(app)['grades']

    def sizing(self, app):
        return self.data(app)['sizing']

    def critical_rules(self, app):
        return self.data(app)['critical_rules']

    def tqi_compliance(self, app):
        return self.data(app)['tqi_compliance']

    def doc_compliance(self, app):
        return self.data(app)['doc_compliance']

    def aggregate_violation_ratio(self,domain_id, snapshot_id,key,sub_keys,crit_only=True):
        self.debug(f'aggregating violation ration information for {domain_id}{snapshot_id}')
        # sid = self.snapshot(app)['id']
        # domain_id = self.domain(app)
        
        df = DataFrame(columns=['Key','Rule','Detail','Weight','Total','Failed','Succeeded','Compliance','Score'])
        indicators = self.get_quality_indicators(domain_id,snapshot_id, key)

        for ind in indicators:
            ind_key = ind['key']
            if ind_key in sub_keys:
                ind_name = ind['name'] #.replace('-','\n  ',1)
                ind_weight = ind['weight']
                vr = self.get_violation_ratio(domain_id,ind_key)
                try:
                    if vr:
                        temp = json_normalize(vr)
                        if crit_only:
                            temp = temp[temp['reference.critical']==True]
                        agr = temp.agg({ 
                            'result.violationRatio.totalChecks': ['sum'], 
                            'result.violationRatio.failedChecks': ['sum'], 
                            'result.violationRatio.successfulChecks': ['sum'] 
                            })
        #                agr['compliance']=(agr['result.violationRatio.successfulChecks']/agr['result.violationRatio.totalChecks'])*100
                        grade_set=self.get_grade(domain_id,ind_key)
                        grade=round(grade_set[0]['result']['grade'],2)

                        compliance=grade_set[0]['result']['score']*100

                        # temp = temp[temp['result.grade']<=2]
                        # Get Top 3 rules, based on max weight of Failed check logic  
                        temp = temp.sort_values('result.violationRatio.failedChecks', ascending = False).head(3)
                        
                        detail = "\n".join(temp['reference.name'].tolist())
                        total=int(agr.iloc[0]['result.violationRatio.totalChecks'])
                        if total == 0:
                            continue
                        failed=int(agr.iloc[0]['result.violationRatio.failedChecks'])
                        success=int(agr.iloc[0]['result.violationRatio.successfulChecks'])

                        df.loc[len(df)] = [
                            ind_key,
                            ind_name,
                            detail,
                            ind_weight,
                            f'{total:,}',
                            f'{failed:,}',
                            f'{success:,}',
                            f'{int(compliance)}%',
                            grade
                        ]                
                except (KeyError):
                    self.logger.warn(f'Key Error while aggregating {ind_name}')

        # for ind in indicators:
        #     key = ind['key']
        #     name = ind['name']
        #     weight = ind['weight']
        #     vr = self.get_violation_ratio(domain_id,key)
        #     if vr:
        #         temp = json_normalize(vr)
        #         temp['key']=key
        #         temp['name']=name
        #         temp['weight']=weight
        #         df = concat([df,temp])

        # df.rename(columns={"name":"parent","reference.name":"name","reference.critical":"critical"},inplace=True)
        # df.rename(columns={"result.grade":"grade","result.violationRatio.totalChecks":"totalChecks"},inplace=True)
        # df.rename(columns={"result.violationRatio.failedChecks":"failedChecks"},inplace=True)
        # df.rename(columns={"result.violationRatio.successfulChecks":"successfulChecks"},inplace=True)
        # df.rename(columns={"result.violationRatio.violationOccurrences":"violationOccurrences"},inplace=True)
        # df.rename(columns={"result.violationRatio.ratio":"ratio"},inplace=True)

        # return df[[ 'type','key', 'parent', 'weight','name','critical', 'grade','totalChecks', 'failedChecks', \
        #             'violationOccurrences', 'successfulChecks','ratio', 'result.score']]

        return df

    def action_plan(self, app):
        ap_df = self.data(app)['action_plan']
        ap_summary_df = self.data(app)['action_plan_summary']
        return (ap_df,ap_summary_df)
        
    def get_app_grades(self, app, sort=False):
        app_grades = self.grades(app).loc['All']
        if sort:
            return app_grades
        else:
            return app_grades.sort_values()

    def calc_grades_all_apps(self):
        all_app=DataFrame()
        for row in self._data:
            if self.has_data(row):
                app_name=self.snapshot(row)['name']
                grades=self.grades(row)
                all_app = concat([all_app,grades[grades.index.isin(['All'])].rename(index={'All': app_name})]).drop_duplicates()
        return all_app[all_app.columns].mean(axis=0)

    def calc_grades_health(self,grade_all):
        grade_df = DataFrame(grade_all)
        grade_health=grade_df[grade_df.index.isin(self._health_grade_ids)]
        return grade_health

    def calc_health_grades_high_risk(self,grade_all):
        grade_health = self.calc_grades_health(grade_all)
        grade_at_risk=grade_health[grade_health < 2].dropna()
        return grade_at_risk

    def calc_health_grades_medium_risk(self,grade_all):
        grade_health = self.calc_grades_health(grade_all)
        grade_at_risk=grade_health[grade_health > 2].dropna()
        grade_at_risk=grade_health[grade_health < 3].dropna()
        return grade_at_risk

    def get_loc_sizing(self,app):
        return self.data(app)['loc_sizing']

    def tech_sizing(self, app):
        return self.data(app)['tech_sizing']

    def violation_sizing(self, app):
        return self.data(app)['violation_sizing']

    def get_all_app_text(self):
        rslt = ""

        data = self._data
        l = len(self._base)
        if l == 1:
            return self.snapshot(self._base[0])['name']

        last_name = self._base[-2]
        for a in self._base:
            if self.has_data(a):
                rslt = rslt + self.snapshot(a)['name']
                if l >= 2 and a == last_name:
                    rslt = rslt + " and "
                elif a != self._base[-1]:
                    rslt = rslt + ", "
            else:
                rslt = "NO SNAPSHOT INFORMATION AVAILABLE"
        return rslt

    def get_grade_by_tech(self,app):
        grade_df = self.grades(app).round(2).applymap('{:,.2f}'.format)
        grade_df = self.grades(app).round(2)
        grade_df = grade_df[grade_df.index.isin(['All'])==False]

        sizing_df = DataFrame(self.sizing(app))
        sizing_df = sizing_df[sizing_df.index.isin(['All'])==False]
        sizing_df = DataFrame(sizing_df["Number of Code Lines"].rename("LOC")).dropna()
        sizing_df.sort_values(by=['LOC'], ascending=False,inplace=True)
        sizing_df = sizing_df.applymap('{:,.0f}'.format)
        
        tech = sizing_df.join(grade_df) 

        sizing_df = DataFrame(self.sizing(app)) 
        sizing_df = sizing_df[sizing_df.index.isin(['All'])==False]
        sizing_df = DataFrame(sizing_df["Critical Violations"]).dropna()
        sizing_df = sizing_df.applymap('{:,.0f}'.format)

        tech = tech.join(sizing_df)
#        tech.sort_values(by='LOC',ascending=False,inplace=True)
#        tech = tech.applymap('{:,.2f}'.format)

        
        return tech

    def get_high_risk_grade_text(self, grades):
        grade_at_risk=self.calc_health_grades_high_risk(grades)
        if grade_at_risk.empty:
            return None
        else:
            return self.text_from_list(grade_at_risk.index.values.tolist())

    def get_medium_risk_grade_text(self, grades):
        grade_at_risk=self.calc_health_grades_medium_risk(grades)
        if grade_at_risk.empty:
            return None
        else:
            return self.text_from_list(grade_at_risk.index.values.tolist())

    def text_from_list(self,list):
        rslt = ""
        total_items = len(list)
        if total_items == 1:
            return list[0]

        last_item = list[-1]
        sec_last_item = list[-2]
        for name in list:
            rslt = rslt + name
            if total_items >= 2 and name == sec_last_item:
                rslt = rslt + " and "
            elif name != last_item:
                rslt = rslt + ", "
        return rslt



class HLData(HLRestCall):
    _data={}

    def __init__(self, config, timer_on=False,log_level=INFO):
        super().__init__(hl_base_url=f'{config.hl_url}/WS2/', hl_user=config.hl_user, hl_pswd=config.hl_password,
                         hl_instance=config.hl_instance, timer_on=timer_on,log_level=log_level)

        for s in config.hl_list:
            self.info(f'Collecting Highlight data for {s}')
            self._data[s]={}
            self._data[s]['has data'] = False

            app_id = self.get_app_id(s)
            if app_id:
                (lic,cves,total_components) = self.get_third_party(app_id)
                
                self._data[s]['has data'] = True
                self._data[s]['app_id']=app_id
                self._data[s]['cves']=cves
                self._data[s]['licenses']=lic
                self._data[s]['cloud']=self.get_cloud_data(app_id)

                if cves is None:
                    self._data[s]['cve_crit_tot']=0
                    self._data[s]['cve_crit_comp_tot']=0
                    self._data[s]['cve_high_tot']=0
                    self._data[s]['cve_high_comp_tot']=0
                    self._data[s]['cve_med_tot']=0
                    self._data[s]['cve_med_comp_tot']=0
                else:
                    self._data[s]['cve_crit_tot']=len(cves[cves['criticity']=='CRITICAL']['cve'].unique())
                    self._data[s]['cve_crit_comp_tot']=len(cves[cves['criticity']=='CRITICAL']['component'].unique())
                    self._data[s]['cve_high_tot']=len(cves[cves['criticity']=='HIGH']['cve'].unique())
                    self._data[s]['cve_high_comp_tot']=len(cves[cves['criticity']=='HIGH']['component'].unique())
                    self._data[s]['cve_med_tot']=len(cves[cves['criticity']=='MEDIUM']['cve'].unique())
                    self._data[s]['cve_med_comp_tot']=len(cves[cves['criticity']=='MEDIUM']['component'].unique())
                
                if lic is None:
                    self._data[s]['lic_high_tot']=0
                    self._data[s]['lic_med_tot']=0
                else:
                    # make some adjustment since HL data has changed
                    lic['compliance']=lic['compliance'].str.replace('compliant','high')
                    lic['compliance']=lic['compliance'].str.replace('partial','medium')
                    lic['compliance']=lic['compliance'].str.replace('notCompliant','low')

                    self._data[s]['lic_high_tot']=len(lic[lic['compliance']=='low']['component'].unique())
                    self._data[s]['lic_med_tot']=len(lic[lic['compliance']=='medium']['component'].unique())

                self._data[s]['total_components']=total_components
            else:
                self.error(f'Highlight Application Id not found for {s}')    
        #    self._base = schema
            self._data_retrieved = False
            self._app_id = None
            self._got_data = False
            self._has_crit_sev_cves = False
            self._has_high_sev_cves = False
            self._has_med_sev_cves = False
            self._has_high_risk_lics = False
            self._has_med_risk_lics = False
            self._third_party_df = DataFrame()
            self._cve_df = DataFrame()
            self._lic_df = DataFrame()

    def get_cve_data(self,app_id):
        return self._data[app_id]['cves']

    def has_data(self,app_id):
        return self._data[app_id]['has data'] 

    def get_data(self,app_id):
        return self._data[app_id]

    def get_cloud_info(self,app_id):
        return self._data[app_id]['cloud'] 

    def get_cve_crit_tot(self,app_id):
        return self._data[app_id]['cve_crit_tot']
    def get_cve_crit_comp_tot(self,app_id):
        return self._data[app_id]['cve_crit_comp_tot']

    def get_cve_high_tot(self,app_id):
        return self._data[app_id]['cve_high_tot']
    def get_cve_high_comp_tot(self,app_id):
        return self._data[app_id]['cve_high_comp_tot']

    def get_cve_med_tot(self,app_id):
        return self._data[app_id]['cve_med_tot']
    def get_cve_med_comp_tot(self,app_id):
        return self._data[app_id]['cve_med_comp_tot']

    def get_lic_high_tot(self,app_id):
        return self._data[app_id]['lic_high_tot']

    def get_lic_med_tot(self,app_id):
        return self._data[app_id]['lic_med_tot']

    def get_oss_cmpn_tot(self,app_id):
        return self._data[app_id]['total_components']
        

    def get_third_party_info(self,app_id):
        return DataFrame(self._data[app_id]['components'])

    def get_lic_info(self,app_name):
        """
            Extract all license relavent columns from the components DF 
        """
        lic = self._data[app_name]['licenses']

        #adjust license risk factors
        if lic is None:
            lic = DataFrame()
        else:
            lic.loc[lic['compliance']=='high','risk']='Low'
            lic.loc[lic['compliance']=='medium','risk']='Medium'
            lic.loc[lic['compliance']=='low','risk']='High'
            lic.loc[lic['compliance']=='undefined','risk']='Undefined'
       
        return lic

    def get_cve_info(self,app_name):
        """
            Extract all license relavent columns from the components DF 
        """
        cves = self._data[app_name]['cves']
        oss_df = DataFrame(columns=['component','critical','high','medium'])

        if cves is not None and not cves.empty:

            oss_work_df = cves[['component','cve','criticity']].copy()
            oss_work_df.drop_duplicates(['component','cve','criticity'],inplace=True)
            oss_work_df.sort_values(by=['component'],inplace=True)

            cve_critical_df = oss_work_df[oss_work_df['criticity']=='CRITICAL']
            cve_critical_df = cve_critical_df.groupby('component')['cve'].apply(', '.join).reset_index()
            cve_critical_df.rename(columns={'cve':'critical'},inplace=True)

            cve_high_df = oss_work_df[oss_work_df['criticity']=='HIGH']
            cve_high_df = cve_high_df.groupby('component')['cve'].apply(', '.join).reset_index()
            cve_high_df.rename(columns={'cve':'high'},inplace=True)

            cve_medium_df = oss_work_df[oss_work_df['criticity']=='MEDIUM']
            cve_medium_df = cve_medium_df.groupby('component')['cve'].apply(', '.join).reset_index()
            cve_medium_df.rename(columns={'cve':'medium'},inplace=True)

            oss_df = cve_critical_df.merge(cve_high_df,on='component',how='outer')
    #        oss_df = oss_df.merge(cve_high_df,on='component',how='outer')
            oss_df = oss_df.merge(cve_medium_df,on='component',how='outer')
            oss_df = oss_df.where(notnull(oss_df),'')
            oss_df = oss_df[['component','critical','high','medium']]

        return oss_df



        # lic = pd.json_normalize(components,
        #                         record_path =['licenses'],
        #                         meta=['componentId','name','version','release',
        #                               'languages','lastVersion','lastRelease',
        #                               'nbVersionPreviousYear'],
        #                         meta_prefix='comp.',
        #                         errors='ignore')

        # lic=lic[['comp.componentId', 'comp.name', 'comp.version',
        #     'comp.release', 'comp.languages', 'comp.lastVersion',
        #     'comp.lastRelease', 'comp.nbVersionPreviousYear','name', 'compliance']]

        # lic.rename(columns={'name':'lic.name','compliance':'compliance'},inplace=True)

        # #adjust license risk factors
        # lic.loc[lic['compliance']=='high','compliance']='Low'
        # lic.loc[lic['compliance']=='medium','compliance']='Medium'
        
        # return lic

    def sort_lic_info(self, lic_df):
        if lic_df is None:
            return lic_df
        else:
            lic_all = lic_df
            try:
                lic_all.sort_values(by=['component','release'],inplace=True)
                lic_high = lic_all[lic_all['compliance']=="High"]
                lic_medium = lic_all[lic_all['compliance']=="Medium"]
                return concat([lic_high,lic_medium])
            except (KeyError):
                return lic_df


    def get_app_ids(self, instance_id):
        # TODO: try-except
        return self._get_app_ids(instance_id)

    def _get_third_party(self, app_id):

        if self._app_id != app_id:
            self._app_id = app_id
            self._got_data = False

        if not self._got_data:
            # If we do not have the data already for this app, retrieve it first.

            self._third_party_df = DataFrame(self.get_third_party(app_id)['thirdParties'])
            # TODO: Confirm that data was retrieved before setting this to True
            self._got_data = True

            self._cve_df = self._third_party_df.loc[:, ['name', 'cve']]
            self._cve_df.dropna(axis = 0, how = 'any', inplace = True)
            self._lic_df = self._third_party_df.loc[:, ['name', 'licenses']]
            self._lic_df.dropna(axis = 0, how = 'all', inplace = True)

        return

    def get_cves(self, components, type, limit = 0):
        """
        Returns CVE info for a give app, if there are Critical, High and medium CVEs.
        Unless the 'all' argument is provided, High CVEs are returned only when there are no/not enough critical CVEs
        to fill the slide. Similarly, medium CVEs are returned only where there are no/limited critical or high CVEs.

        Note that low CVEs are ignored for assessment purposes.
        """

        _cve_df = DataFrame()
        sev_type = type.lower()

        # If the request is for critical sev CVEs, override the limit and return all rows.
        if sev_type == 'critical':
            limit = 0

        # Do we have the data retrieved for the app? If not, auto-retrieve.
        # If all == False, limit the number of CVEs returned.

        # try:
        #     self._get_third_party(app_id)
        # except:
        #     print('ERROR - no thirdparty data')
        #     raise

        cve_df = components.loc[:, ['name', 'cve']]
        cve_df.dropna(axis = 0, how = 'any', inplace = True)


        i = 0
        prev_comp = ''
        comp_changed = False

        for tup in self._cve_df.itertuples():
            crit_cve_str, high_cve_str, med_cve_str = '', '', ''
            comp_name = tup[1]

            for cve in tup[2]['vulnerabilities']:
                # If critical sev is the type requested, also grab high and medium sev CVEs for the same component.
                # Similarly, if high sev is the type requested, also grab medium sev CVEs for the same component.
                criticality = cve['criticity'].lower()

                # Ignore components with low criticality, as we do not need to report those.
                if criticality == 'low':
                    continue

                # Note that are not considering the type requested here, but simply storing the CVE in its
                # appropriate slot. That part is handled further below, before they are added to the dataframe.
                if criticality == 'critical':
                    crit_cve_str += cve['name'] + ', '
                if criticality == 'high':
                    high_cve_str += cve['name'] + ', '
                elif criticality == 'medium':
                    med_cve_str += cve['name'] + ', '

            # Before adding the results into the dataframe, ensure that we have matches for the
            # crticality requested. For example, if critical CVEs were requested, we should have
            # values in the crit_cve_str. # If not, do not add.
            if sev_type == 'critical' and len(crit_cve_str) > 0 or \
                sev_type == 'high' and len(high_cve_str) > 0    or \
                sev_type == 'medium' and len(med_cve_str) > 0:

                # Strip the comma-spaces at the end of the CVE strings.
                crit_cve_str = crit_cve_str[:-2]
                high_cve_str = high_cve_str[:-2]
                med_cve_str = med_cve_str[:-2]

                if len(crit_cve_str) == 0:
                    crit_cve_str = 'N/A'
                if len(high_cve_str) == 0:
                    high_cve_str = 'N/A'
                if len(med_cve_str) == 0:
                    med_cve_str = 'N/A'

                _cve_df = _cve_df.append({'Component Name': comp_name, 'Critical Sev CVEs': crit_cve_str,
                            'High Sev CVEs': high_cve_str, 'Medium Sev CVEs': med_cve_str}, ignore_index = True)

            i += 1

            # Return only the max number of rows requested. In case the sev type is critical, always return all rows.
            if (limit > 0) and (i == limit):
                break

        #print(_cve_df)
        return _cve_df

    def get_lics(self, app_id, type = 'high', limit = 0):
        """
        """

        compliance = ''
        _lic = []
        _lic_df = DataFrame()

        # Do we have the data retrieved for the app? If not, auto-retrieve.
        # If all == False, limit the number of CVEs returned.

        try:
            self._get_third_party(app_id)
        except:
            print('ERROR - no thirdparty data')
            raise

        i = 0
        for tup in self._lic_df.itertuples():
            high_risk_lic_str = ''
            med_risk_lic_str = ''
            comp_name = tup[1]

            for dic in tup[2]:
                if type == 'high':
                    # A note on risk-complaince relation.
                    # A riskier license translates to lower compliance. The REST API returns compliance info.
                    # So, if the type being sought is high-risk license, we return low compliance license.
                    #
                    # If tbe requested type is high risk licenses and we also see
                    # medium risk licenses for the same component, include those as well.
                    # Leveraging the content of high_risk_lic_str for this purpose.

                    if dic['compliance'] == 'low':
                        high_risk_lic_str += dic['name'] + '\n'
                    elif dic['compliance'] == 'medium' and len(high_risk_lic_str) > 0:
                        med_risk_lic_str += dic['name'] + '\n'
                elif type == 'medium':
                    # A medium risk likcense has a compliance of medium.
                    if dic['compliance'] == 'medium': 
                        med_risk_lic_str += dic['name'] + '\n'
            
            if len(high_risk_lic_str) > 0 or len(med_risk_lic_str) > 0:
                # Strip the trailing newline.
                high_risk_lic_str = high_risk_lic_str[:-1]
                med_risk_lic_str = med_risk_lic_str[:-1]

                if len(high_risk_lic_str) == 0:
                    high_risk_lic_str = 'N/A'

                if len(med_risk_lic_str) == 0:
                    med_risk_lic_str = 'N/A'

                _lic_df = _lic_df.append({'Component Name': comp_name, 'High Risk': high_risk_lic_str,
                            'Medium Risk': med_risk_lic_str}, ignore_index = True)
                i += 1

            if (limit > 0) and (i == limit):
                break

        """
        for elem in _lics:
            print(elem.key, elem.value)
        for idx in range(len(_lics)):
            for key in _lics[idx]:
                print(key, _lics[idx][key])

            #for j in range(len(high_risk_lic_df['LicenseInfo'][i])):
                #print(high_risk_lic_df['LicenseInfo'][i][j])
                #_lic_str += high_risk_lic_df['name'][i][j] + '\n'

            #_lic_str = _lic_str[:-1]
            #_lics = _lics.append({'name:', high_risk_lic_df['licenses'], 'licenses:', _lic_str})
        """
        #print(_lic_df)
        return _lic_df

# rest = AipData('http://sha-dd-node02:8080/CAST-RESTAPI-integrated/rest/','cast','cast',['MainWebApplication'],log_level=DEBUG)
# vr = rest.aggregate_violation_ratio('MainWebApplication',"60017")



from keycloak import KeycloakAdmin
from cast_common.aipRestCall import AipRestCall

keycloak_admin = KeycloakAdmin(server_url="http://arch-ps-2:8086/auth/",
                               username='admin',
                               password='admin',
                               verify=True)
user = keycloak_admin.get_users()

for item in user:
    if item['username']=='admin':
        user_id = item['id']
        break

clients = keycloak_admin.get_clients()
for item in clients:
    client_id = item['id']
    roles = keycloak_admin.get_available_client_roles_of_user(user_id,client_id)

pass



from cast_arg.restCall import AipData,HLData
#from cast_arg.powerpoint import PowerPoint
from cast_arg.actionPlan import ActionPlan
from cast_arg.config import Config
from cast_arg.pages.greenIt import GreenIt
from cast_arg.pages.summary import HighlightSummary
from cast_arg.stats import OssStats,AIPStats,LicenseStats

from cast_arg.pages.tech_detail_table import TechDetailTable


from cast_common.mri import MRI
from cast_common.logger import Logger,DEBUG, INFO, WARN
from cast_common.util import find_nth, no_dups, list_to_text,format_table
from cast_arg.powerpoint import PowerPoint
from cast_common.highlight import Highlight


from pandas import DataFrame
from pptx import Presentation
from pptx.dml.color import RGBColor
from IPython.display import display
from os import getcwd
from os.path import abspath,dirname,exists
from site import getsitepackages

import pandas as pd
import numpy as np 
import math
import json
import datetime

__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2022, CAST Software"

class GeneratePPT(Logger):
    _ppt = None
    _aip_data = None
    _effort_df = None
    _output_folder = None
    _project_name = None
    _template = None

    _generate_HL = _hl_base_url = _hl_user = hl_pswd = None
    _generate_AIP = _aip_base_url = _aip_user = _aip_pswd = None
    _hl_instance = None
    _hl_apps_df = pd.DataFrame()
    _hl_app_list = []

    day_rate=1600

    def __init__(self, config:Config):
        super().__init__("generate",config.logging_generate)
        self._config = config

        out = f"{config.output}/Project {config.project} - Tech DD Findings.pptx"
        self.info(f'Generating {out}')

        self._ppt = PowerPoint(config.template, out)

        # TODO: Handle cases where on HL data is needed and not AIP.
        self.hl_pages = []
        if config.aip_active:
            self.info("Collecting AIP Data")
            self._aip_data = AipData(config,log_level=config.logging_aip)

            self.mri_pages = [
                TechDetailTable()

            ]

        if config.hl_active:
            self.info("Collecting Highlight Data")
            hl = Highlight(hl_base_url=config.hl_url,hl_user=config.hl_user,hl_pswd=config.hl_password, \
                           hl_instance=config.hl_instance,hl_apps=config.hl_list)
            self.hl_pages = [
                HighlightSummary(self.day_rate),
                GreenIt(self._config.output)
            ]
            self._hl_data = HLData(config,log_level=config.logging_highlight)

        #project level work
        app_cnt = len(config.application)

        # self.remove_proc_slides(self._generate_procs)

        self._ppt.duplicate_slides(app_cnt)
        self._ppt.copy_block("each_app",["app"],app_cnt)

        self._ppt.replace_text("{app#_","{app1_")

        self.replace_all_text()

    def remove_proc_slides(self,keep_it):
        indexes=[]
        for idx, slide in enumerate(self._ppt._prs.slides):
            for shape in slide.shapes:
                if shape.has_text_frame:
                    for paragraph in shape.text_frame.paragraphs:
                        if paragraph.text == "{proc_analysis}":
                            self._ppt.replace_slide_text(slide,"{proc_analysis}","")
                            indexes.append(idx)

        for idx in reversed(indexes):
            if not keep_it:
                self._ppt.delete_slide(idx)

    def save_ppt(self):
        self._ppt.save()

    def replace_all_text(self):
        app_cnt = len(self._config.application)
        self._ppt.replace_text("{project}", self._config.project)
        self._ppt.replace_text("{app_count}",app_cnt)
        self._ppt.replace_text("{company}",self._config.company)

        mydate = datetime.datetime.now()
        month = mydate.strftime("%B")
        year = f'{mydate.year} '
        self._ppt.replace_text("{month}",month)
        self._ppt.replace_text("{year}",year)

        # replace AIP data global to all applications
        if self._config.aip_active:
            all_apps_avg_grade = self._aip_data.calc_grades_all_apps()
#            self._ppt.replace_text("{all_apps}",self._aip_data.get_all_app_text())
            self._ppt.replace_risk_factor(all_apps_avg_grade,search_str="{summary_")
            risk_grades = self._aip_data.calc_health_grades_high_risk(all_apps_avg_grade)
            if risk_grades.empty:
                risk_grades = self._aip_data.calc_health_grades_medium_risk(all_apps_avg_grade)
            self._ppt.replace_text("{summary_at_risk_factors}",list_to_text(risk_grades.index.tolist()).lower())
        
            if risk_grades.empty:
                self._ppt.replace_block('{immediate_action}','{end_immediate_action}','')
            else:
                self._ppt.replace_text("{immediate_action}","") 
                self._ppt.replace_text("{end_immediate_action}","") 
                self._ppt.replace_text("{high_risk_grade_names}",self._aip_data.text_from_list(risk_grades.index.values.tolist()))

        app_list = self._config.aip_list
        hl_list = self._config.hl_list

        day_rate = self.day_rate

        summary_near_term=AIPStats(day_rate,logger_level=self._config.logging_generate)
        summary_fix_now=AIPStats(day_rate,logger_level=self._config.logging_generate)
        summary_mid_long_term=AIPStats(day_rate,logger_level=self._config.logging_generate)
        hl_summary=OssStats('',day_rate,logger_level=self._config.logging_generate)
        hl_summary_critical=OssStats('',day_rate,logger_level=self._config.logging_generate)
        hl_summary_high_near=OssStats('',day_rate,logger_level=self._config.logging_generate)
        lic_summary = LicenseStats(logger_level=self._config.logging_generate)
        summary_components = 0

        self._ppt.replace_text('{app_cnt}',app_cnt)
        for idx in range(0,app_cnt):
            # create instance of action plan class 
            self.ap = ActionPlan (app_list,self._aip_data,self._ppt,self._config.output,day_rate)
            fix_now_total=AIPStats(day_rate,logger_level=self._config.logging_generate)
            near_term_total=AIPStats(day_rate,logger_level=self._config.logging_generate)
            mid_long_term=AIPStats(day_rate,logger_level=self._config.logging_generate)
            summary_total=AIPStats(day_rate,logger_level=self._config.logging_generate)

            hl_near_term_total=AIPStats(day_rate,logger_level=self._config.logging_generate)

            # replace application specific AIP data
            app_no = idx+1
            app_id = app_list[idx]
            hl_id = hl_list[idx]
            app_title = self._config.title_list[idx]
            self.info(f'********************* Working on pages for {app_title} ******************************')
            self._ppt.replace_text(f'{{app{app_no}_name}}',app_title)

            if self._config.aip_active:
                for proc in self.mri_pages:
                    proc.report(app_id,app_no)

            if self._config.hl_active:
                for proc in self.hl_pages:
                    proc.report(hl_id,app_no)

            if self._config.aip_active:
                if self._aip_data.has_data(app_id):
                    self.info('Preparing AIP Data')
                    #self.info(f'Working on {app_id} ({self._appl_title[app_id]})')

                    self.info('Filling risk factors for the executive summary page')
                    # do risk factors for the executive summary page
                    self.fill_aip_grades(self._aip_data,app_id, app_no)
                    risk_grades = self.each_risk_factor(self._aip_data,app_id, app_no)
                    self._ppt.replace_text(f'{{app{app_no}_high_risk_grade_names}}',list_to_text(risk_grades.index.values))


                    self.info('Filling Technical details TABLE')
                    # Technical Overview - Technical details TABLE
                    grade_all = self._aip_data.get_app_grades(app_id)
                    #self._ppt.replace_risk_factor(grade_all,app_no)
                    grade_by_tech_df = self._aip_data.get_grade_by_tech(app_id)
                    # grades = grade_by_tech_df.drop(['Documentation',"ISO","ISO_EFF","ISO_MAINT","ISO_REL","ISO_SEC"],axis=1)
                    # self._ppt.update_table(f'app{app_no}_grade_by_tech_table',grades)

                    if not grade_by_tech_df.empty:
                        #add appmarq technology
                        self._ppt.replace_text(f'{{app{app_no}_largest_tech}}',grade_by_tech_df.index[0])

                        self.info('Filling Technical Overview')
                        # Technical Overview - Lines of code by technology GRAPH
                        self._ppt.update_chart(f'app{app_no}_sizing_pie_chart',DataFrame(grade_by_tech_df['LOC']))

                    snapshot = self._aip_data.snapshot(app=app_id)
                    # app_name = snapshot['name']
                    # self._ppt.replace_text(f'{{app{app_no}_name}}',app_name)
                    self._ppt.replace_text(f'{{app{app_no}_all_technogies}}',list_to_text(snapshot['technology']))

                    #calculate high and medium risk factors
                    risk_grades = self._aip_data.calc_health_grades_high_risk(grade_all)
                    if risk_grades.empty:
                        risk_grades = self._aip_data.calc_health_grades_medium_risk(grade_all)
                    self._ppt.replace_text(f'{{app{app_no}_at_risk_grade_names}}',list_to_text(risk_grades.index.tolist()).lower())

                    self.fill_strengh_improvement_tbl(app_id,app_no)
                    
                    """
                        Populate the document insites page
                        The necessary data is found in the loc_tbl

                        This section fetches data for Documentation slide, excludes certain columns, sorts with few column, 
                        and based on score of particular element seggregates colors(Red, Yellow & Green) and update to app1_doc_table element.
                        255,168,168 - Red shades
                        255,234,168 - Yellow shades
                        168,228,195 - Green shades
                    """
                    doc_df = self._aip_data.doc_compliance(app_id)
                    doc_df.drop(columns=['Key','Total','Weight'],inplace=True) #'Detail',
                    doc_df.sort_values(by=['Score','Rule'], inplace=True)
                    doc_df['RGB'] = np.where(doc_df.Score >= 3,'194,236,213',np.where(doc_df.Score < 2,'255,210,210','255,240,194'))
                    doc_df.Score = doc_df.Score.map('{:.2f}'.format)
                    self._ppt.update_table(f'app{app_no}_doc_table',doc_df,include_index=False,background='RGB')
                    


                   
                    loc_df = self._aip_data.get_loc_sizing(app_id)
                    if len(loc_df) > 0:
                        loc = loc_df['Number of Code Lines']
                        self._ppt.replace_loc(loc,app_no)

                        loc_tbl = pd.DataFrame.from_dict(data=self._aip_data.get_loc_sizing(app_id),orient='index').drop('Critical Violations')
                        loc_tbl = loc_tbl.rename(columns={0:'loc'})
                        loc_tbl['percent'] = round((loc_tbl['loc'] / loc_tbl['loc'].sum()) * 100,2)
                        loc_tbl['loc']=pd.Series(["{0:,.0f}".format(val) for val in loc_tbl['loc']], index = loc_tbl.index)

                        percent_comment = loc_tbl.loc['Number of Comment Lines','percent']
                        percent_comment_out = loc_tbl.loc['Number of Commented-out Code Lines','percent']

                        if percent_comment < 15:
                            comment_level='low'
                        elif percent_comment > 15 and percent_comment <= 20:
                            comment_level='good'
                        else:
                            comment_level='high'
                    
                        self._ppt.replace_text(f'{{app{app_no}_comment_hl}}',comment_level)
                        self._ppt.replace_text(f'{{app{app_no}_comment_level}}',comment_level)
                        self._ppt.replace_text(f'{{app{app_no}_comment_pct}}',percent_comment)
                        self._ppt.replace_text(f'{{app{app_no}_comment_out_pct}}',percent_comment_out)

                        loc_tbl['percent']=pd.Series(["{0:.2f}%".format(val) for val in loc_tbl['percent']], index = loc_tbl.index)
                        self._ppt.update_table(f'app{app_no}_loc_table',loc_tbl,header_rows=0)
                        self._ppt.update_chart(f'app{app_no}_loc_pie_chart',DataFrame(loc_tbl['loc']))

                        # self._ppt.replace_grade(grade_all,app_no)

                        self.fill_sizing(app_id,app_no)
                        self.fill_violations(app_id,app_no)
    
                    self.fill_critical_rules(app_id,app_no)

                    """
                        Get Action Plan data and combine it for the various slide combinations

                        Executive Summary uses a combination of extreme and high violation data
                        Action Plan midigation slice is divided into three sections
                            1. Fix before sigining contains only extreme violation data
                            2. Near term contains high violation data
                            3. Long term contains medium and low violation data
                            4. Excecutive Summary contains fix now and high violation data
                    """
                    self.ap.fill_action_plan(app_id,app_no)
                    fix_now_total.add_effort(self.ap.fix_now.effort)
                    fix_now_total.add_violations(self.ap.fix_now.violations)

                    near_term_total.add_effort(self.ap.high.effort)
                    near_term_total.add_violations(self.ap.high.violations)

                    mid_long_term.add_effort(self.ap.medium.effort)
                    mid_long_term.add_violations(self.ap.medium.violations)
                    mid_long_term.add_data(self.ap.medium.data)
                    mid_long_term.add_effort(self.ap.low.effort)
                    mid_long_term.add_violations(self.ap.low.violations)
                    mid_long_term.add_data(self.ap.low.data)

                    # summary_near_term.add_effort(self.ap.high.effort)
                    # summary_near_term.add_effort(self.ap.medium.effort)
                    # summary_near_term.add_violations(self.ap.high.violations)
                    # summary_near_term.add_violations(self.ap.medium.violations)

                    """
                        ISO-5055 slide
                            - most of the work is being done in the restCall class
                            - use iso_rules to retrieve the data
                            - add it to the app1_iso5055 table in the template
                    """
                    iso_df = self._aip_data.iso_rules(app_id)
                    if not iso_df.empty:
                        iso_df.loc[iso_df['violation']=='','background']='205,218,226'
                        iso_df.loc[iso_df['violation']!='','background']='255,255,255'
                        self._ppt.update_table(f'app{app_no}_iso5055',iso_df,
                                            include_index=False,background='background')
                                       
                        pourcentage_iso5055 = iso_df["count"].sum()
                        
                        iso_Maintainaility = iso_df[iso_df.category == 'Maintainability' ]
                        iso_MaintainailityCall = iso_Maintainaility["count"].sum()
                        self._ppt.replace_text(f'{{app{app_no}_ISO_5055}}', round((iso_MaintainailityCall/(pourcentage_iso5055/2))*100,1))
                    
            #replaceHighlight application specific data
            if self._config.hl_active and self._hl_data.has_data(hl_id):
                try:
                    (oss_crit,oss_high,oss_med,lic,components) = self.oss_risk_assessment(hl_id,app_no,day_rate)
                    fix_now_total.add_effort(oss_crit.effort)
                    fix_now_total.add_violations(oss_crit.violations)

                    near_term_total.add_effort(oss_high.effort)
                    near_term_total.add_effort(oss_med.effort)

                    hl_near_term_total.add_effort(oss_high.effort)
                    hl_near_term_total.add_effort(oss_med.effort)
                    hl_near_term_total.add_violations(oss_high.violations)
                    hl_near_term_total.add_violations(oss_med.violations)
                    hl_near_term_total.replace_text(self._ppt,app_no,'hl_near_term_total')

                    hl_summary_critical.add_components(oss_crit.components)
                    hl_summary_critical.add_violations(oss_crit.violations)

                    hl_summary_high_near.add_components(oss_high.components)
                    hl_summary_high_near.add_components(oss_med.components)

                    hl_summary.add_components(components)

#                    if not lic.empty:
                    lic_summary.add_high(lic.high)
                    lic_summary.add_medium(lic.medium)
                    lic_summary.add_low(lic.low)


                except KeyError as ex:
                    self.warning(f'OSS information not found {str(ex)}')
                """
                    Cloud ready excel sheet generation
                """
                try:
                    cloud = self._hl_data.get_cloud_info(hl_id)
                    cloud = cloud[['cloudRequirement.display','Technology','cloudRequirement.ruleType','cloudRequirement.criticality','contributionScore','roadblocks']]
                    file_name = f'{self._config.output}/cloud-{self._config.title_list[idx]}.xlsx'
                    writer = pd.ExcelWriter(file_name, engine='xlsxwriter')
                    col_widths=[50,10,10,10,10,10,10]
                    cloud_tab = format_table(writer,cloud,'Cloud Data',col_widths)
                    writer.close()
                except Exception as e:
                    self.error(f'unknown error while processing cloud ready data: {str(e)}')


            summary_fix_now.add_effort(fix_now_total.effort)
            summary_fix_now.add_violations(fix_now_total.violations)
            summary_near_term.add_effort(near_term_total.effort)
            summary_near_term.add_violations(near_term_total.violations)
            summary_mid_long_term.add_effort(mid_long_term.effort)
            summary_mid_long_term.add_violations(mid_long_term.violations)

            summary_total.add_effort(summary_fix_now.effort)
            summary_total.add_effort(summary_near_term.effort)
            summary_total.add_effort(summary_mid_long_term.effort)
            summary_total.add_violations(summary_fix_now.violations)
            summary_total.add_violations(summary_near_term.violations)
            summary_total.add_violations(summary_mid_long_term.violations)
            summary_total.add_data(summary_fix_now.data)
            summary_total.add_data(summary_near_term.data)
            summary_total.add_data(summary_mid_long_term.data)
            summary_total.replace_text(self._ppt,app_no,'summary_total')


            #replace text for all aip and HL statistics in the powerpoint document
            self.ap.fix_now.replace_text(self._ppt,app_no,'fix_now')
            self.ap.high.replace_text(self._ppt,app_no,'high')
            self.ap.medium.replace_text(self._ppt,app_no,'medium')
            self.ap.low.replace_text(self._ppt,app_no,'low')

            mid_long_term.replace_text(self._ppt,app_no,'mid_long_term')

            fix_now_total.replace_text(self._ppt,app_no,'fix_now_total')
            near_term_total.replace_text(self._ppt,app_no,'near_term_total')

            # short_term.replace_text(self._ppt,app_no,'st')
            # long_term.replace_text(self._ppt,app_no,'lt')
        
        #summary_fix_now_aip_hl.replace_text(self._ppt,'','hl_summary_fix_now')
        summary_fix_now.replace_text(self._ppt,'','summary_fix_now')
        summary_near_term.replace_text(self._ppt,'','summary_near_term')

        hl_summary_critical.replace_text(self._ppt,'_summary_crit')
        hl_summary_high_near.replace_text(self._ppt,'_summary_high_near')
        lic_summary.replace_text(self._ppt,'_summary')
        hl_summary.replace_text(self._ppt,'_summary')
        self._ppt.replace_text('{app_summary_comp_tot}',summary_components)

        show_stopper_flg = 'some'
        if self.ap.fix_now.effort == 0:
            show_stopper_flg = 'no'
        self._ppt.replace_text('{show_stopper_flg}',show_stopper_flg)  

        avg_cost = float(summary_fix_now.cost/app_cnt)   
        if avg_cost < 50:
            hml_cost_flg = 'low'
        elif avg_cost > 50 and avg_cost < 100:
            hml_cost_flg = 'medium'
        else:
            hml_cost_flg = 'high'
        self._ppt.replace_text('{hml_cost_flag}',hml_cost_flg)  

        self._ppt.replace_text('{daily_rate}',self.ap._day_rate)  

    def each_risk_factor(self, aip_data, app_id, app_no):
        # collect the high risk grades
        # if there are no high ris grades then use medium risk
        app_level_grades = aip_data.get_app_grades(app_id)
        risk_grades = aip_data.calc_health_grades_high_risk(app_level_grades)
        risk_catagory="high"
        if risk_grades.empty:
            risk_catagory="medium"
            risk_grades = aip_data.calc_health_grades_medium_risk(app_level_grades)
        
        #in the event all health risk factors are low risk
        if risk_grades.empty:
            self._ppt.replace_block(f'{{app{app_no}_risk_detail}}',
                            f'{{end_app{app_no}_risk_detail}}',
                            "no high-risk health factors")
        else: 
            rpl_str = f'{{app{app_no}_risk_category}}'
            self._ppt.replace_text(rpl_str,risk_catagory)
            self.debug(f'replaced {rpl_str} with {risk_catagory}')

            self._ppt.copy_block(f'app{app_no}_each_risk_factor',["_risk_name","_risk_grade"],len(risk_grades.count(axis=1)))
            f=1
            for index, row in risk_grades.T.items():
                rpl_str = f'{{app{app_no}_risk_name{f}}}'
                self._ppt.replace_text(rpl_str,index)
                self.debug(f'replaced {rpl_str} with {index}')

                rpl_str = f'{{app{app_no}_risk_grade{f}}}'
                value = row['All'].round(2)
                self._ppt.replace_text(rpl_str,value)
                self.debug(f'replaced {rpl_str} with {value}')
                f=f+1

        self._ppt.remove_empty_placeholders()
        return risk_grades

    def get_grade_color(self,grade):
        rgb = 0
        if grade > 3:
            rgb = RGBColor(0,176,80) # light green
        elif grade <3 and grade > 2:
            rgb = RGBColor(214,142,48) # yellow
        else:
            rgb = RGBColor(255,0,0) # red
        return rgb

    def fill_aip_grades(self,aip_data, app_id, app_no):
        self.info('Filling AIP grades data')
        app_level_grades = aip_data.get_app_grades(app_id)
        for name, value in app_level_grades.T.items():
            # fill grades
            grade = round(value,2)
            rpl_str = f'{{app{app_no}_grade_{name}}}'
            self._ppt.replace_text(rpl_str,grade)
            self.debug(f'replaced {rpl_str} with {grade}')

            # fill grade risk factor (high, medium or low)
            rpl_str = f'{{app{app_no}_risk_{name}}}'
            risk = ''
            if grade < 2:
                risk = 'high'
            elif grade < 3:
                risk = 'medium'
            else:
                risk = 'low'
            self._ppt.replace_text(rpl_str,risk)
            self.debug(f'replaced {rpl_str} with {risk}')

            #update grade box color and slider postion 
            id_base = f'app{app_no}_grade'
            box_name = f'{id_base}_{name}_box'
            txt_name = f'{id_base}_{name}_text'
            slider_name = f'{id_base}_{name}_slider'
            color = self.get_grade_color(grade)

            for slide in self._ppt._prs.slides:
                box = self._ppt.get_shape_by_name(box_name,slide)
                if not box is None:
                    box.line.color.rgb = color

                txt = self._ppt.get_shape_by_name(txt_name,slide)
                if not txt is None and txt.has_text_frame:
                    paragraphs = txt.text_frame.paragraphs
                    self._ppt.change_paragraph_color(paragraphs[0],color)

                slider = self._ppt.get_shape_by_name(slider_name,slide)
                if not slider is None:
                    self._ppt.update_grade_slider(slider,[grade])




    def fill_strengh_improvement_tbl(self,app_id,app_no):
        self.info('Filling strength and improvment table')
        """
            Populate the strengths and improvement page
            The necessary data is found in the loc_tbl
        """
        imp_df = self._aip_data.tqi_compliance(app_id)
        imp_df.drop(columns=['Weight','Total','Succeeded','Compliance'],inplace=True)
        imp_df.sort_values(by=['Score','Rule'], inplace=True, ascending = False)

        file_name = f'{self._config.output}/health-{self._config.title_list[app_no-1]}.xlsx'
        writer = pd.ExcelWriter(file_name, engine='xlsxwriter')
        col_widths=[50,50,10,10,10]
        cloud_tab = format_table(writer,imp_df,'Health Data',col_widths)
        writer.close()

        imp_df.drop(columns=['Detail'],inplace=True)
        imp_df['RGB'] = np.where(imp_df.Score >= 3,'194,236,213',\
            np.where(imp_df.Score < 2,'255,210,210','255,240,194'))
        imp_df.Score = imp_df.Score.map('{:.2f}'.format)

        #cause_name = abspath(f'{dirname(__file__)}/cause.json')

        cause_name = abspath(f'{getsitepackages()[-1]}/cast_arg/cause.json')
        imp_df['Cause']=''
        with open(cause_name) as json_file:
            tech_data = json.load(json_file)
        imp_df['Cause']=imp_df['Key'].map(tech_data)

        imp_df = imp_df[['Rule','Score','Cause','Failed','RGB']]
        self._ppt.update_table(f'app{app_no}_imp_table',imp_df,include_index=False,background='RGB')

    def fill_critical_rules(self,app_id,app_no):
        self.info('Filling critical rules table')
        rules_df = self._aip_data.critical_rules(app_id)
        if not rules_df.empty:
            rules_df = rules_df[['rulePattern.name','rulePattern.critical']]
            rule_summary_df=rules_df.groupby(['rulePattern.name']).size().reset_index(name='counts').sort_values(by=['counts'],ascending=False)
            rule_summary_df=rule_summary_df.head(5)
            self._ppt.update_table(f'app{app_no}_top_violations',rule_summary_df,include_index=False)
        else:
            self.warning('This application contains no critical violations')
        # if not rules_df.empty:
        #     critical_rule_df = pd.json_normalize(rules_df['rulePattern'])
        #     critical_rule_df = critical_rule_df[['name','critical']]

        #     #pourcentage_iso5055 = critical_rule_df['name']
        #     #self._ppt.replace_text(f'{{app{app_no}_ISO_5055}}',rule_summary_df)

    def fill_violations(self,app_id,app_no):
        self.info('Filling violation table')
        violation_df = pd.DataFrame(self._aip_data.violation_sizing(app_id),index=[0])
        violation_df['Violation Count']=pd.Series(["{0:,.0f}".format(val) for val in violation_df['Violation Count']])
        violation_df[' per file']=pd.Series(["{0:,.2f}".format(val) for val in violation_df[' per file']])
        violation_df[' per kLoC']=pd.Series(["{0:,.2f}".format(val) for val in violation_df[' per kLoC']])
        violation_df['Complex objects']=pd.Series(["{0:,.0f}".format(val) for val in violation_df['Complex objects']])
        violation_df[' With violations']=pd.Series(["{0:,.0f}".format(val) for val in violation_df[' With violations']])
        self._ppt.update_table(f'app{app_no}_violation_sizing',violation_df.transpose())
        self._ppt.replace_text(f'{{app{app_no}_critical_violations}}',violation_df['Violation Count'].loc[0])

        #violation_df['Security']=pd.Series(["{0:,.0f}".format(val) for val in violation_df['Security']])
        
        #print('#########################################################################################')
        #print('coucou le resultat est',violation_df['Security'])
        #print('#########################################################################################')
        ######################################################################

        ##code added for is 5055
        #pourcentage_iso5055 = violation_df[' per file'].sum()
        #self._ppt.replace_text(f'{{app{app_no}_ISO_5055}}',violation_df['Violation Count'].sum())
        ######################################################################################
        

    def fill_sizing(self,app_id,app_no):
        self.info('Filling sizing table')
        sizing_df = pd.DataFrame(self._aip_data.tech_sizing(app_id),index=[0])
        sizing_df['LoC']=pd.Series(["{0:,.0f} K".format(val / 1000) for val in sizing_df['LoC']])
        sizing_df['Files']=pd.Series(["{0:,.0f}".format(val) for val in sizing_df['Files']])
        sizing_df['Classes']=pd.Series(["{0:,.0f}".format(val) for val in sizing_df['Classes']])
        sizing_df['SQL Artifacts']=pd.Series(["{0:,.0f}".format(val) for val in sizing_df['SQL Artifacts']])
        sizing_df['Tables']=pd.Series(["{0:,.0f}".format(val) for val in sizing_df['Tables']])
        sizing_df = sizing_df.transpose()
        self._ppt.update_table(f'app{app_no}_tech_sizing',sizing_df)

    def oss_risk_assessment(self,hl_id,app_no,day_rate):
        self.info('Filling OSS risk assessment table')
        lic_df=self._hl_data.get_lic_info(hl_id)
        lic_df=self._hl_data.sort_lic_info(lic_df)
        oss_df=self._hl_data.get_cve_info(hl_id)
        # lic_summary = pd.DataFrame(columns=['License Type','Risk Factor','Component Count','Example'])

        oss_crit = OssStats(hl_id,day_rate,self._hl_data,'crit',logger_level=self._config.logging_generate)
        oss_crit_comp_tot = self._hl_data.get_cve_crit_comp_tot(hl_id)

        oss_high = OssStats(hl_id,day_rate,self._hl_data,'high',logger_level=self._config.logging_generate)
        oss_high_comp_tot = self._hl_data.get_cve_high_comp_tot(hl_id)

        oss_med = OssStats(hl_id,day_rate,self._hl_data,'med',logger_level=self._config.logging_generate)
        oss_med_comp_tot = self._hl_data.get_cve_med_comp_tot(hl_id)

        total_components = self._hl_data.get_oss_cmpn_tot(hl_id)

        oss_crit.replace_text(self._ppt,app_no)
        oss_high.replace_text(self._ppt,app_no)
        oss_med.replace_text(self._ppt,app_no)

        self._ppt.replace_text(f'{{app{app_no}_oss_cmpn_tot}}',total_components)
        if not oss_df.empty:
            self._ppt.update_table(f'app{app_no}_HL_table_CVEs',oss_df,include_index=False)

        self.info('Filling OSS license table')
        '''
            License compliance table
        
        '''
        lic_summary=self._hl_data.get_lic_info(hl_id)
        if not lic_summary.empty:
            lic_summary=lic_summary[['component','version','release','license','risk']].drop_duplicates()
            lic_summary.sort_values(['risk','license','component','version'],inplace=True)

            lic_summary = lic_summary.groupby(['risk','license'])['component'].apply(lambda x: ','.join(x)).reset_index()
            lic_summary['comp count']=lic_summary['component'].str.count(',')+1
            
            #remove duplicates but show full count
            lic_summary['component']=lic_summary['component'].map(lambda x: no_dups(x,',',True))
            
            #only show the first 5 components
            lic_summary['component']=lic_summary['component'].map(lambda x: x[:find_nth(x,',',6)])
            lic_summary=lic_summary[['license','risk','comp count','component']]
            lic_summary.sort_values(['risk','license','comp count'],inplace=True)

            #remove low and undefined records from the table
            lic_summary=lic_summary[lic_summary["risk"].str.contains("Low")==False]
            lic_summary=lic_summary[lic_summary["risk"].str.contains("Undefined")==False]

            # are there any records left?
            lic = LicenseStats(logger_level=self._config.logging_generate)
            if not lic_summary.empty:
                #modify the forground color
                lic_summary.loc[lic_summary['risk']=='High','forground']='211,76,76'
                lic_summary.loc[lic_summary['risk']=='Medium','forground']='127,127,127'

                #update the powerpoint table
                self._ppt.update_table(f'app{app_no}_HL_table_lic_risks',lic_summary,include_index=False)
            
                lic.high = lic_summary[lic_summary['risk']=='High']['comp count'].sum()
                lic.medium = lic_summary[lic_summary['risk']=='Medium']['comp count'].sum()
                lic.low = lic_summary[lic_summary['risk']=='Low']['comp count'].sum()

                # #add the high and medium license risk counts to the deck
                # self._ppt.replace_text(f'{{app{app_no}_high_lic_tot}}',
                #     lic_summary[lic_summary['risk']=='High']['comp count'].sum())
                # self._ppt.replace_text(f'{{app{app_no}_med_lic_tot}}',
                #     lic_summary[lic_summary['risk']=='Medium']['comp count'].sum())

            lic.replace_text(self._ppt,app_no)
        else:
            lic = DataFrame()
        return (oss_crit,oss_high,oss_med,lic,total_components)








"""
    Open Source Safty Statistical Information
"""
from pandas import DataFrame, concat
from cast_common.logger import Logger,INFO
from cast_common.util import list_to_text,rreplace
import math

__author__ = "Nevin Kaplan"
__copyright__ = "Copyright 2022, CAST Software"
__email__ = "n.kaplan@castsoftware.com"



class AIPStats():
    def __init__(self,day_rate,logger_level=INFO):
        self._log = Logger('AIPStats',logger_level)

        self._day_rate=day_rate
        self._effort=0
        self._violations=0
        self._data = DataFrame()

    @property
    def effort(self):
        return self._effort
    @effort.setter
    def effort(self,value):
        self._effort = int(value)
    def add_effort(self,value):
        self._effort = self._effort + int(value)
    def sub_effort(self,value):
        self._effort = self._effort - int(value)


    @property
    def cost(self):
        return self.effort * self._day_rate /1000

    @property
    def violations(self):
        return self._violations
    @violations.setter
    def violations(self,value):
        self._violations = int(value)
    def add_violations(self,value):
        self._violations = self._violations + int(value)
    def sub_violations(self,value):
        self._violations = self._violations - int(value)

    @property
    def data(self) -> DataFrame:
        return self._data
    @data.setter
    def data(self,value):
        self._data = value
    def add_data(self,data):
        self.data = concat([self.data,data],ignore_index=True)

    def business_criteria(self,filtered):
        list = []
        if not filtered.empty:
            for business in filtered['Business Criteria']:
                items = business.split(',')
                for t in items:
                    if t.strip() not in list:
                        list.append(t.strip())
        return list

    def list_violations(self,filtered):
        first = True
        text = ""
        try:
            for criteria in filtered['Technical Criteria'].unique():
                df = filtered[filtered['Technical Criteria']==criteria]
                total = df['No. of Actions'].sum()
                
                cases = 'for'
                if first:
                    cases = 'cases of'
                    first = False
                
                rule = criteria[criteria.find('-')+1:].strip().lower()
                if len(rule) == 0:
                    rule = criteria
                text = f'{text}{total} {cases} {rule}, '
            return rreplace(text[:-2],', ',' and ')
        except (KeyError):
            return ""

    def replace_text(self,ppt,app_no,priority):
        bus_txt = list_to_text(self.business_criteria(self.data)) + ' '
        vio_txt = self.list_violations(self.data)
        base_txt = f'app{app_no}_aip_{priority}'
        msg = f'Replacing {base_txt} (_eff={self.effort}, _cost={self.cost}, _vio_cnt={self.violations}'

        ppt.replace_text(f'{{{base_txt}_eff}}',self.effort)
        ppt.replace_text(f'{{{base_txt}_cost}}',self.cost)
        ppt.replace_text(f'{{{base_txt}_vio_cnt}}',self.violations)
        if len(bus_txt) > 0:
            msg = f'{msg}, bus_txt={bus_txt}'
            ppt.replace_text(f'{{{base_txt}_bus_txt}}',bus_txt)
        if len(vio_txt) > 0:
            msg = f'{msg}, vio_txt={vio_txt}'
            ppt.replace_text(f'{{{base_txt}_vio_txt}}',vio_txt)
        msg = f'{msg})'
        self._log.debug(msg)
        

class LicenseStats():
    def __init__(self,logger_level=INFO):
        self._log = Logger('LicenseStats',logger_level)

        self._high=0
        self._medium=0
        self._low=0

    @property
    def high(self):
        return self._high
    @high.setter
    def high(self,value):
        self._high = int(value)
    def add_high(self,value):
        self._high = self._high + int(value)

    @property
    def medium(self):
        return self._medium
    @medium.setter
    def medium(self,value):
        self._medium = int(value)
    def add_medium(self,value):
        self._medium = self._medium + int(value)

    @property
    def low(self):
        return self._low
    @low.setter
    def low(self,value):
        self._low = int(value)
    def add_low(self,value):
        self._low = self._low + int(value)

    def replace_text(self,ppt,app_no):
        base_txt = f'app{app_no}_hl_lic'
        self._log.debug(f'Replacing {base_txt} tags (_high={self.high}, _med={self.medium}, _low={self.low})')
        ppt.replace_text(f'{{{base_txt}_high}}',self.high)
        ppt.replace_text(f'{{{base_txt}_med}}',self.medium)
        ppt.replace_text(f'{{{base_txt}_low}}',self.low)



class OssStats():
    __priorities = ['crit','high','med']

    def __init__(self, app_id=None, day_rate=-1, data=None, priority=None,logger_level=INFO):
        # super().__init__("CVE")
        self._log = Logger('OssStats',logger_level)

        if app_id is None:
            raise KeyError('invalid parameters: Highlight Application Id is required')
        if day_rate <= 0:
            raise KeyError('invalid parameters: day rate must be greater than zero')
        
        if priority is None:
            priority = ''
        else:
            if priority not in self.__priorities:
                txt = ', '.join(self.__priorities)
                raise KeyError(f'invalid parameters: Priority must be {txt}')

        self._priority = priority
        self._day_rate = day_rate

        if data is None:
            self._violations = 0
            self._components = 0
            self._effort = 0
        else:
            self._violations = data.get_data(app_id)[f'cve_{self._priority}_tot']
            self._components = data.get_data(app_id)[f'cve_{self._priority}_comp_tot']

    @property
    def violations(self):
        return int(self._violations)
    @violations.setter
    def violations(self,value):
        self._violations = int(value)
    def add_violations(self,value):
        self._violations = self._violations + int(value)

    @property
    def components(self):
        return self._components
    @components.setter
    def components(self,value):
        self._components = int(value)
    def add_components(self,value):
        self._components = self._components + int(value)

    @property
    def effort(self):
        return math.ceil(self._components/2)

    @property
    def cost(self):
        return self.effort * self._day_rate /1000

    def replace_text(self,ppt,app_no):
        base_txt = f'app{app_no}_hl_{self._priority}'
        self._log.debug(f'Replacing {base_txt} tags (_eff={self.effort}, _comp_tot={self.components}, _cost={self.cost}, _vio_cnt={self.violations})')
        ppt.replace_text(f'{{{base_txt}_eff}}',self.effort)
        ppt.replace_text(f'{{{base_txt}_comp_tot}}',self.components)
        ppt.replace_text(f'{{{base_txt}_cost}}',self.cost)
        ppt.replace_text(f'{{{base_txt}_vio_cnt}}',self.violations)





from setuptools import setup,find_packages

setup (
    name="com.castsoftware.uc.arg",
    url='https://github.com/CAST-Extend/com.castsoftware.uc.arg',
    description="Assessment Report Generator",
#    install_requires=['pandas','python-pptx==0.6.18','com.castsoftware.uc.python.common>=0.1.6','IPython','requests','Jinja2'],
    package_data={'':['cause.json','Effort.csv']},
    packages=find_packages()
)



from cast_arg.stats import AIPStats
from cast_common.logger import Logger,INFO
from cast_common.aipRestCall import AipRestCall
from cast_common.util import format_table,list_to_text
from os import getcwd
from os.path import abspath,dirname,exists
from site import getsitepackages


import math
import pandas as pd

__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2022, CAST Software"

"""
    This class is used to collect action plan information and add them to the 
    the proper tags
"""
class ActionPlan(AipRestCall):
    _app_list = []
    _ppt = None
    _aip_data = None
    _effort_df = None
    _output_folder = None

    def __init__(self,app_list,aip_data,ppt,output_folder,day_rate,logger_level=INFO):
        self._app_list = app_list
        self._output_folder=output_folder
        self._ppt = ppt
        self._aip_data=aip_data

        #ef_name = abspath(f'{dirname(__file__)}/Effort.csv')
        ef_name = abspath(f'{getsitepackages()[-1]}/cast_arg/Effort.csv')

        if not exists(ef_name):
            raise RuntimeError(f'Required file not found: {ef_name}')
        self._effort_df = pd.read_csv(ef_name)

        self._day_rate = day_rate
        self._fix_now = AIPStats(day_rate)
        self._high = AIPStats(day_rate)
        self._med = AIPStats(day_rate)
        self._low = AIPStats(day_rate)


    @property
    def fix_now(self): return self._fix_now
    @property
    def high(self): return self._high
    @property
    def medium(self): return self._med
    @property
    def low(self): return self._low
    @property
    def day_rate(self): return self._day_rate


    def fill_action_plan(self,app_id,app_no):

        (ap_df,ap_summary_df)=self._aip_data.action_plan(app_id)
        if not ap_summary_df.empty:
            ap_summary_df = ap_summary_df.merge(self._effort_df, how='inner', on='Technical Criteria')
            #cost_col = (ap_summary_df['Eff Hours'] * ap_summary_df['No. of Actions'])/8
            ap_summary_df['Days Effort'] = (ap_summary_df['Eff Hours'] * ap_summary_df['No. of Actions'])/8
            ap_summary_df['Cost Est.'] = ap_summary_df['Days Effort'] * self._day_rate

            file_name = f'{self._output_folder}/{app_id}_action_plan.xlsx'
            writer = pd.ExcelWriter(file_name, engine='xlsxwriter')
            col_widths=[50,40,10,10,10,50,10,10,10]
            summary_tab = format_table(writer,ap_summary_df[['Quality Rule','Business Criteria','No. of Actions','comment']],'Summary',col_widths)
            col_widths=[10,50,50,30,30,30,30,30,30,30,30,30,30]
            format_table(writer,ap_df,'Action Plan',col_widths)
            writer.close()

            #fill action plan related tags
            self._fix_now = self.calc_action_plan_effort(ap_summary_df,app_no,'extreme','security')
            self._high = self.calc_action_plan_effort(ap_summary_df,app_no,'high')
            self._med = self.calc_action_plan_effort(ap_summary_df,app_no,'moderate')
            self._low = self.calc_action_plan_effort(ap_summary_df,app_no,'low')
            

            #configure action plan table background colors 
            ap_summary_df.loc[ap_summary_df['tag']=='extreme','RGB']='244,212,212'
            ap_summary_df.loc[ap_summary_df['tag']=='high','RGB']='255,229,194'
            ap_summary_df.loc[ap_summary_df['tag']=='moderate','RGB']='203,225,238'
            ap_summary_df.loc[ap_summary_df['tag']=='low','RGB']='254,254,255'

            ap_table = pd.concat([ap_summary_df[ap_summary_df['tag']=='extreme'],
                                  ap_summary_df[ap_summary_df['tag']=='high'],
                                  ap_summary_df[ap_summary_df['tag']=='moderate'],
                                  ap_summary_df[ap_summary_df['tag']=='low']])

            ap_table = ap_table.drop(columns=['comment','tag','Technical Criteria','Days Effort','Cost Est.','Eff Hours'])

            self._ppt.update_table(f'app{app_no}_action_plan',ap_table.head(29),include_index=False,background='RGB')

            sum = ap_summary_df['No. of Actions'].sum()
            self._ppt.replace_text(f"{{app{app_no}_total_violations}}",str(sum))

            violation_table = self._ppt.get_shape_by_name(f'app{app_no}_action_plan') 
            if violation_table:
                page_no = self._ppt.get_page_no(violation_table)
                self._ppt.replace_text(f"{{app{app_no}_violation_page}}",str(page_no))

            violation_table = self._ppt.get_shape_by_name(f'app{app_no}_table_type_name') 
            if violation_table:
                page_no = self._ppt.get_page_no(violation_table)
                self._ppt.replace_text(f"{{app{app_no}_HL_violation_page}}",str(page_no))
        else:
            self._ppt.replace_text(f"{{app{app_no}_extreme_violation_total}}",'TBD') 
            self._ppt.replace_text(f"{{app{app_no}_high_violation_total}}",'TBD') 
            self._ppt.replace_text(f"{{app{app_no}_moderate_violation_total}}",'TBD') 
            self._ppt.replace_text(f"{{app{app_no}_low_violation_total}}",'TBD') 

    def calc_action_plan_effort(self,ap_summary_df,app_no,priority,default='') -> AIPStats: 
        rslt = AIPStats(self._day_rate)
        (priority_text, vio_cnt, rslt.data) = self.common_business_criteria(ap_summary_df,priority,default)
        rslt.effort = math.ceil(rslt.data['Days Effort'].sum()*2)
        rslt.violations = int(vio_cnt)
        return rslt

    # def fill_action_plan_tags(self,app_no,type,effort,cost,vio_cnt,bus_txt,vio_txt):
    #     self._ppt.replace_text(f'{{app{app_no}_{type}_eff}}',effort)
    #     self._ppt.replace_text(f'{{app{app_no}_{type}_cost}}',cost)
    #     self._ppt.replace_text(f'{{app{app_no}_{type}_vio_cnt}}',vio_cnt)

    #     self._ppt.replace_text(f'{{app{app_no}_{type}_bus_txt}}',bus_txt,tbd_for_blanks=False)
    #     self._ppt.replace_text(f'{{app{app_no}_{type}_vio_txt}}',vio_txt)

    def common_business_criteria(self,summary_df,priority,default=''):
        filtered=summary_df[summary_df['tag']==priority]
        count = 0
        sum = 0
        list = []
        if not filtered.empty:
            sum = filtered['No. of Actions'].sum()
            list = self.business_criteria(filtered)
        sum_txt = str(sum)
        
        if not list:
            list.append(default)

        return list_to_text(list),sum_txt, filtered

    def business_criteria(self,filtered):
        list = []
        if not filtered.empty:
            for business in filtered['Business Criteria']:
                items = business.split(',')
                for t in items:
                    if t.strip() not in list:
                        list.append(t.strip())
        return list


    # def list_violations(self,filtered):
    #     first = True
    #     text = ""
    #     try:
    #         for criteria in filtered['Technical Criteria'].unique():
    #             df = filtered[filtered['Technical Criteria']==criteria]
    #             total = df['No. of Actions'].sum()
                
    #             cases = 'for'
    #             if first:
    #                 cases = 'cases of'
    #                 first = False
                
    #             rule = criteria[criteria.find('-')+1:].strip().lower()
    #             if len(rule) == 0:
    #                 rule = criteria
    #             text = f'{text}{total} {cases} {rule}, '
    #         return util.rreplace(text[:-2],', ',' and ')
    #     except (KeyError):
    #         return ""
            




from cast_common.highlight import Highlight
from cast_common.logger import Logger, INFO,DEBUG
from cast_common.powerpoint import PowerPoint
from cast_common.util import list_to_text,convert_LOC

from pandas import json_normalize

class HighlightSummary(Highlight):

    def __init__(self,day_rate:int):
        super().__init__()
        self._day_rate = day_rate

    def report(self,app_name:str,app_no:int) -> bool:
        app_tag = f'app{app_no}'

        #create list of technolgies sorted by LOC in decending order
        tech_list = list_to_text(self.get_technology(app_name)['technology'].to_list())
        self.replace_text(app_tag,'tech',tech_list)

        loc=self.get_total_lines_of_code(app_name)
        (total_loc,unit) = convert_LOC(int(loc))
        self.replace_text(app_tag,'total_loc',f'{total_loc} {unit}')
        
        comp_total = self.get_component_total(app_name)
        self.replace_text(app_tag,'component_total',f'{comp_total:,}')

        critical = len(self.get_cve_critical(app_name)['cve'].unique())
        self.replace_text(app_tag,'cve_critical_total',f'{critical:,}')

        high = len(self.get_license_high(app_name)['component'].unique())
        self.replace_text(app_tag,'license_high_total',f'{high:,}')

        cost = ((critical * self._day_rate)/2)/1000
        self.replace_text(app_tag,'oss_cost',f'{cost:,.1f}')

        self.replace_text(app_tag,'software_health_hml',self.get_software_health_hml(app_name).capitalize())
        self.replace_text(app_tag,'elegance_score',f'{self.get_software_elegance_score(app_name):,.1f}')
        self.replace_text(app_tag,'agility_score',f'{self.get_software_agility_score(app_name):,.1f}')
        self.replace_text(app_tag,'resiliency_score',f'{self.get_software_resiliency_score(app_name):,.1f}')

        cloud_ready = self.get_cloud_detail(app_name)
        blockers = len(cloud_ready[cloud_ready['cloudRequirement.ruleType']=='BLOCKER'])
        self.replace_text(app_tag,'cloud_ready_blocker_total',f"{blockers:,f}")

        pass

    def replace_text(self,prefix, item, data):
        tag = f'{{{prefix}_{item}}}'
        self.log.debug(f'{tag}: {data}')
        PowerPoint.ppt.replace_text(tag,data)

# from os.path import abspath
# from cast_common.util import format_table
# from pandas import ExcelWriter

# ppt = PowerPoint(r'E:\work\Decks\highlight-test.pptx',r'E:\work\Decks\test\highlight.pptx')

# app = 'CollabServer'
                            
# hl = HighlightSummary('n.kaplan+insightsoftwareMinerva@castsoftware.com','vadKpBFAZ8KIKb2f2y',hl_instance=383,hl_base_url='https://app.casthighlight.com',log_level=DEBUG)
# hl.report(app,1,ppt,r'E:\work\Decks\test')
# ppt.save()

